/*
 * Modul für MMKS Einspritzventil
 * 
 * Version: 1.0 
 * Date: 26.02.2021
 * Author: Christoph - www.inovalab.de
 */ 

#include <avr/wdt.h>
#include <EEPROM.h>
#include <SoftwareSerial.h>
#include "ModbusRtu.h"


/* Arduino Definitions */
const uint8_t ANALOGINPUT = 0;
const uint8_t PWMOUTPUT = 1;

/* Pin Mapping Einspritzventil */
const uint8_t pinAiTon = 21;
const uint8_t pinAiToff = 20;
const uint8_t pinDiCh1 = 17;
const uint8_t pinDiCh2 = 16;
const uint8_t pinOutCh1 = 9;
const uint8_t pinOutCh2 = 10;

void presetPinsFunctions()
{  /* Pin configuration:
    bit 0: 0 keine Funktion
    bit 1: 1 Eingang
    bit 2: 2 Eingang mit Pullup
    bit 3: 3 Analogeingang
    bit 4: 4 Ausgang
    bit 5: 5 PWM Ausgang
  */
  setPinFunction(pinAiTon, 3, 0);
  setPinFunction(pinAiToff, 3, 0);
  setPinFunction(pinDiCh1, 1, 0);
  setPinFunction(pinDiCh2, 1, 0);
  setPinFunction(pinOutCh1, 4, 0);
  setPinFunction(pinOutCh2, 4, 0);
}

uint8_t inputPins[2] = {pinDiCh1, pinDiCh2};

uint8_t pinInCh1State = LOW;
uint8_t pinInCh2State = LOW;

uint8_t pinOutCh1State = LOW;
uint8_t pinOutCh2State = LOW;

/* Einspritzventil Setup */
float ton = 0;
uint16_t tonStart = 500; // 0,5ms, Poti linker Anschlag
uint16_t tonEnd = 10000; // 10ms, Poti rechter Anschlag
uint8_t tonDeadZone = 50;
float tonM = (float)(tonEnd - tonStart)/(1023-tonDeadZone);
uint16_t mReg_ton = 30;

float toff = 0;
uint16_t toffStart = 500; // 0,5ms, Poti linker Anschlag
uint32_t toffEnd = 300000; // 300ms, Poti rechter Anschlag
uint8_t toffDeadZone = 50;
float toffM = (float)(toffEnd - toffStart)/(1023-toffDeadZone);
uint16_t mReg_toff = 1000;

uint32_t previousTime = 0;
uint8_t tState = LOW;


/* Modbus Setup */
const uint8_t pinRx = 7;
const uint8_t pinTx = 4;
const uint8_t pinTxEn = 2;

uint8_t modbus_deviceId = 3; //Modbus Adresse
uint32_t modbus_baudrate = 19200;
SoftwareSerial softSerial(pinRx, pinTx);
Modbus modbusInstance(modbus_deviceId, softSerial, pinTxEn);

// data array for modbus network sharing
const uint8_t au16dataSize = 7;
uint16_t au16data[au16dataSize] = {1, 2, 3, 4, 5, 6, 7};

/* Eprom */
uint16_t checkUpdateValuesEeprom[2];
uint16_t checkUpdateValuesEepromLast[2];

uint16_t writeEpromState = LOW;

uint8_t checkWriteEpromState = LOW;
uint8_t checkWriteEpromStateLast = LOW;


void setup()
{
   //uint8_t epromVal = EEPROM.read(0);
   uint8_t randomNumber = getRandomNumber(1, 254, 0);
   //EEPROM.write(0, randomNumber);
   au16data[0] = randomNumber;
   au16data[1] = 0b00000000; // pinInputState
   au16data[2] = 0b00000000; // pinOutputState
   au16data[3] = 0b00000000; // modbus switching
   au16data[4] = 0b00000000; // ErrorState
   au16data[5] = mReg_ton;
   au16data[6] = mReg_toff;
   
   checkUpdateValuesEeprom[0] = au16data[5];
   checkUpdateValuesEeprom[1] = au16data[6];
   
  
  /* Preconfig */
  presetPinsFunctions(); 
  
  softSerial.begin(modbus_baudrate);//use the hardware serial if you want to connect to your computer via usb cable, etc.
  modbusInstance.start(); // start the ModBus object.
  
  /*Watchdog
  WDTO_15MS   0   15 ms
  WDTO_30MS   1   30 ms
  WDTO_60MS   2   60 ms
  WDTO_120MS  3   120 ms
  WDTO_250MS  4   250 ms
  WDTO_500MS  5   500 ms
  WDTO_1S   6   1 s
  WDTO_2S   7   2 s*/
  wdt_enable(WDTO_120MS);

  Serial.begin(115200);

  /* Eprom lesen */
  uint8_t adresse = 0;
  for ( uint8_t i = 0; i < (sizeof(checkUpdateValuesEeprom) / sizeof(uint16_t)); i++ )
  {
    adresse += sizeof(uint16_t);
    uint16_t epromVal = 0;
    EEPROM.get(adresse, epromVal);
    Serial.print("epromVal[");Serial.print(i);Serial.print("]: ");
    Serial.println(epromVal);
    checkUpdateValuesEeprom[i] = epromVal;
  }
  au16data[5] = checkUpdateValuesEeprom[0];
  au16data[6] = checkUpdateValuesEeprom[1];
}

void loop()
{
  wdt_reset();

  /* Modbus register auf veränderung prüfen */
  checkUpdateValuesEepromLast[0] = au16data[5];
  checkUpdateValuesEepromLast[1] = au16data[6];
  for ( uint8_t i = 0; i < (sizeof(checkUpdateValuesEepromLast) / sizeof(uint16_t)); i++ )
  {
    /*Serial.print("checkUpdateValuesEepromLast[");Serial.print(i);Serial.print("]: ");
    Serial.println(checkUpdateValuesEepromLast[i]);
    Serial.print("checkUpdateValuesEeprom[");Serial.print(i);Serial.print("]: ");
    Serial.println(checkUpdateValuesEeprom[i]);*/
    if (checkUpdateValuesEepromLast[i] != checkUpdateValuesEeprom[i])
    {
      checkUpdateValuesEeprom[i] = checkUpdateValuesEepromLast[i];
      //update eprom state
    }
  }
  checkWriteEpromState = bitRead(au16data[3], 7);
  if ( checkWriteEpromStateLast != checkWriteEpromState )
  {
    if ( checkWriteEpromStateLast == 0 && checkWriteEpromState == 1 ) // Flanke 0 -> 1
    {
      //write eprom
      //Serial.println("write eeprom: ");
      writeEpromState = HIGH;
    }
    checkWriteEpromStateLast = checkWriteEpromState;
  }
  /* Eprom schreiben */
  if ( writeEpromState )
  {
    uint8_t adresse = 0;
    for ( uint8_t i = 0; i < (sizeof(checkUpdateValuesEeprom) / sizeof(uint16_t)); i++ )
    {
     
      adresse += sizeof(uint16_t);
      uint16_t epromVal = 0;
      EEPROM.get(adresse, epromVal);
      if ( epromVal != checkUpdateValuesEeprom[i] )
      {
        EEPROM.put(adresse, checkUpdateValuesEeprom[i]);
        //Serial.print("EEPROM.put[");Serial.print(adresse);Serial.print("]: ");
        //Serial.println(checkUpdateValuesEeprom[i]);
      }
    }
    writeEpromState = LOW;
  }

  /* Eingänge lesen */
  pinInCh1State = digitalRead(pinDiCh1);
  pinInCh2State = digitalRead(pinDiCh2);

  /*  Eingänge in Modbus register schreiben*/
  for ( uint8_t i = 0; i < (sizeof(inputPins) / sizeof(uint8_t)); i++ )
  {
    uint8_t inputState = digitalRead(inputPins[i]);
    if (inputState)
    {
      bitSet(au16data[1], i);
    }
    else
    {
      bitClear(au16data[1], i);
    }
  }

  
  /* Analogeingänge lesen */
  uint16_t rawAiTon = analogRead(pinAiTon);
    //Serial.print("rawAiTon: ");
    //Serial.println(rawAiTon);
  if ( rawAiTon >= tonDeadZone )
  {
    ton = (tonM * (rawAiTon-tonDeadZone) + tonStart);
    /*Serial.print("ton: ");
    Serial.println(ton);
    Serial.print("tonM: ");
    Serial.println(tonM);
    Serial.print("tonStart: ");
    Serial.println(tonStart);*/
  }
  else
  {
    ton = 0;
  }

  uint16_t rawAiToff = analogRead(pinAiToff);
  if ( rawAiToff >= toffDeadZone )
  {
    toff = (toffM * (rawAiToff-toffDeadZone) + toffStart);
    /*Serial.print("toff: ");
    Serial.println(toff);
    Serial.print("toffM: ");
    Serial.println(toffM);
    Serial.print("toffStart: ");
    Serial.println(toffStart);*/
  }
    else
  {
    toff = 0;
  }

  /* Einschalt- Ausschaltzeit */
  uint32_t currentTime = micros();
  if ( HIGH )
  {
    float localTon = ton;
    float localToff = toff;
    if ( bitRead(au16data[3], 0) )
    {
      localTon = (float)au16data[5]*100;
      localToff = (float)au16data[6]*100;
    }
    /*else if ( bitRead(au16data[3], 1) )
    {
      localTon = ton;
      localToff = toff;
    }*/
    /*Serial.print("mReg_ton: ");
    Serial.println(au16data[5]);
    Serial.print("mReg_toff: ");
    Serial.println(au16data[6]);*/
    if ( localTon > 0 &&  localToff > 0 )
    {
      if ( currentTime - previousTime >= localToff )
      {
        previousTime = currentTime;
        tState = HIGH;
        if ( bitRead(au16data[3], 0) || bitRead(au16data[3], 1) || pinInCh1State )
        {
          pinOutCh1State = HIGH;
        }
      }
      if ( tState && currentTime - previousTime >= localTon )
      {
        pinOutCh1State = LOW;
      }
    }
    else
    {
      pinOutCh1State = LOW;
    }
  }

  
  /* Ausgänge setzen*/
  digitalWrite(pinOutCh1, pinOutCh1State);

  
  modbusInstance.poll(au16data, au16dataSize);
  
}

uint8_t getRandomNumber( uint8_t minNumber, uint8_t maxNumber, uint8_t lastNumber )
{
  randomSeed(millis());
  uint8_t number = random(minNumber, maxNumber+1);  // kleinste zahl, größte zahl +1, z.b.: zahl von 1-10: (1,11)
  while (lastNumber == number)
  {
    number = random(minNumber, maxNumber+1);
  }
  return number;
}

void setPinFunction(uint8_t pin, uint8_t pinFunction, uint8_t writeEEprom)
{
  switch (pinFunction)
  {
    case 0: // gesperrt
      break;
    case 1: // Eingang      
      pinMode(pin, INPUT);
      break;
    case 2: // Eingang mit Pullup
      pinMode(pin, INPUT_PULLUP);
      break;
    case 3: // Analogeingang
      pinMode(pin, ANALOGINPUT);
      break;
    case 4: // Ausgang
      pinMode(pin, OUTPUT);
      break;
    case 5: // PWM Ausgang
      pinMode(pin, PWMOUTPUT);
      break;
    default:
      break;
  }  
}
